// matlab.e

// $Revision: 6 $  $Date:: 2012-02-20 09:00:15#$

/* Matlab mode for Epsilon

   This epsilon extension includes auto-indenting, syntax coloring,
   and tagging for The MathWorks Inc. MATLAB(R) source code, version
   7.6 and earlier.

   Matlab mode is automatically invoked on files with extension .m.
   Several variables alter the behavior of matlab mode. They all
   begin with the prefix m-, e.g. m-indent, m-comment-flow, etc. A
   list can be obtained from matlab.h or by using epsilon's variable
   completion on m-. Matlab mode defines a new keytable called m_tab.

   Matlab Auto-indenting
    - The indentation level is controlled by the buffer specific
      variable m-indent. If this variable is zero the current buffer
      tab-size will be used.
    - Indentation occurs everytime the <enter> key is hit, which also
      opens a new line. To indent only, without a new line, use the
      command m-indent-only which is bound to ALT-<enter>
    - When an end, else, otherwise, case, or catch is entered matlab
      mode will reindent the current line. This can be turned off by
      setting m-auto-reindent to zero.
    - The indentation level of the first case statement following a
      switch is specified by the value of the variable m-case-indent.
      If m-case-indent is negative the indent level will be
      determined by m-indent.
    - The command m-indent-region will indent all lines between point
      and mark according to the current indentation settings and rules.
    - Matlab mode includes delimiter matching like epsilon's c-mode.
      It also features end and else statement matching. When the
      keyword "end", "else", or "catch" is entered matlab mode will
      show the matching if, for, while, try or switch statement. Of
      course else will only match an if and catch will only match a
      try. To invoke a match on the most recent previous end, else,
      elseif, or catch use the command m-find-delimiter which is bound
      to Alt-]. If an unmatched end, else, or catch statement is
      entered the bell will sound. This may be turned off by setting 
      m-want-bell to zero. The end matching feature may be turned off
      by setting m-matchdelim to zero; 
    - In-line comments (comments following matlab statements) will be
      indented to the value of m-comment-column if it is non-zero. If
      it is zero the indentation will be to the value of epsilon's
      variable comment-column. To turn off identing of in-line
      comments set m-inline-indent to zero.
    - When entering a comment on a line by itself (full-line) an
      <enter> key will start a new comment on the next line. This may
      be turned off by setting m-comment-flow = 0. If the comment is
      an in-line comment the <enter> key will not generate a new line
      but the indentation will be for a new statement. If m-auto-fill
      is non-zero both full-line and in-line comments will auto-fill
      at the buffer's current fill column.
    - When hitting <enter> on an empty comment line, the line with
      the empty comment will be deleted. To retain the empty comment
      line execute leave-m-comment which is bound to Ctrl-<enter>. To
      always prevent deleting of empty comment lines set
      m-comment-delete-empty = 0. 
    - If the variable m-auto-show-delimiters is non-zero, epsilon's
      auto-show-delimiters feature will operate on the set of
      delimiters defined in m-auto-show-delim-chars. This set can be
      changed. Changing either of the matlab variables requires
      re-invoking matlab-mode before changes will take effect.

   Matlab Syntax coloring
    - As with other epsilon coloring modes, to turn it off set
      want-code-coloring to zero.
    - Matlab coloring introduces new color classes. They are similar to
      the C color classes and pick up the C color_class values by
      default. They are:
          m_comment;         // % this is a comment                                         
          m_keyword;         // for, if, else, end, global, ...                             
          m_class_keyword    // classdef, properties, methods, ...
          m_class_attribute  // classdef attributes - GetAccess, SetAccess, Visible, ...
          m_function;        // a matlab function, e.g. inv(a)                              
          m_identifier;      // variable or object                                          
          m_string;          // 'this is a string'                                          
          m_number;          // 1.5e-10, Inf, Nan, end                                      
          m_punctuation;     // =,*,-,^,.^,.*,[],(),{},;,:,etc                              
          m_unmatched_delim; // { without a }, ] without a [ etc                            
          m_unmatched_keyword; // for without an end, end without a for/while/if/try/switch 
          m_unterminated;    // 'this is an unterminated string                             
      It is generally best to assign function and identifier classes
      the same color since one cannot differentiate between an
      indexed array and a function.
      
   Matlab Tagging
    - Matlab tagging was developed primarily to support long function
      switchyards which are used to write GUIs. It also tags function
      names, including subfunctions. The original motivation was to
      tag constructs such as
         if strcmp(action,'tag1')
           a=1;  ...
         elseif strcmp(action,'tag2')
           a=2;  ...
         elseif findstr(action,'tag3')
           a=3;  ...
           .
           .
         end
      In Matlab 5 this became
         switch action
           case 'tag1'
             a=1; ...
           case 'tag2'
             a=2; ...
           case {'tag3','tag4'}
             .
             .
         end
      Both of these are supported. Currently elseif and case are only
      tagged if they begin in column 0.
    - To tag matlab files (*.m files) run the epsilon command
      tag-files, bound to Ctrl-X Alt-., and provide a file or a
      wildcard specification of a set of .m files. The epsilon
      tag commands such as goto-tag, pluck-tag, etc can now be used.
    - If m-tag-function is non-zero, lines which begin with
         function [x,y,...] = tagident(a,b,...)
      will be tagged with the tag "tagident"
    - If m-tag-case is non-zero, lines which begin with
         case 'tagident'
      will be tagged with the tag "tagident". This version does not
      tag cell array expressions, i.e., case {'hello','goodbye'}
    - If m-tag-elseif is non-zero, lines which begin with
         elseif anyfunc(a,b,..,'tagident',...)
      will be tagged with the tag "tagident". The last string which
      encloses a single identifer on that line will be the tag.
    - The code between two tagged lines is called a section (this
      will depend on the settings of m-tag-function, m-tag-case, and 
      m-tag-elseif). The command m-narrow-to-section on Alt-n will mark
      a section as a region and invoke narrow-to-region. Alt-n is a
      toggle and will widen the buffer if currently narrowing.
    - To move from tagged line to tagged line in the current buffer
      use m-find-next-section on Alt-. and m-find-previous-section on
      Alt-,.

The code in this extension relies heavily on epsilon's c.e and colcode.e
      
Please send any bugs or comments on this matlab mode to:

Terry Brennan
the Optical Sciences Company
tbrennan@tosc.com

*/

/* note: there is an eel bug in find_group when doing a reverse search
         the workaround - replace
           find_group(n,1) with MIN(find_group(n,1),find_group(n,0))
           find_group(n,0) with MAX(find_group(n,1),find_group(n,0))

   note 2: When doing narrowing with coloring some versions of
           epsilon may consume all the cpu time. This was fixed with
           a patch from Lugaru, cpu.e. This may be unnecessary in an
           epsilon version > 9.0. (It is fixed in 9.02)
*/

/* Revision history:
   1.01 Fixed Auto-reindent when at end of buffer 1/23/98
        Added leave-m-comment on Ctrl-<enter> 2/10/98
        Fixed problem with matrix assignment indentation
          (first-matrix) 2/10/98
				Added multiple tagging for case {'option1','option2',...} 2/11/98
				Added conditional compile (EELVERSION>89) for
				   auto_show_matching_characters definition 2/17/98
				Added pause to list of keywords 2/25/98
   1.02 Added try/catch to keyword list 4/13/99
        Added Auto-indent (and re-indent) to try/catch/end 4/19/99
        Fixed a problem which interpretted a % in a string as an inline comment
           and indented it. This required a change to is-in-string.
        Added m_case_indent variable to control indentation level of
           case statement
        Added m-indent-region
   1.03 Removed m-indent-region - fixed m-compute-indent() so that
           epsilon's indent-region does the job for me. 10/18/99
           Epsilon's indent-region will call line by line whatever
           function is bound to the tab key.
        Added variable reindent-after-matlab-yank to work like
           c-mode epsilon's. 10/18/99 That's why indent-region had to be fixed
        Fixed do-m-indent to work like do-c-indent. Added variable
           m-tab-always-indents 10/19/99.
        Eliminated the command m-indent-alone. It's function is replaced by
           do-m-indent. There is a subroutine m-indent-alone which sets up an
           always indent situation and calls do-m-indent. 10/21/99
        Corrected bugs in is_m_(else|case|catch) 10/21/99
   1.1  Added m-comment-next-line, m-uncomment-next-line,
              m-comment-region, and m-uncomment-region 11/9/99
        Added m-cx-tab keytable to support above commands 11/9/99
        Added context menu support for region commenting 11/9/99
        leave-m-comment, m-comment-handling where changed from commands
           to subroutines
        Rename commands:
          find-next-section     --> m-find-next-section
          find-previous-section --> m-find-previous-section
          narrow-to-section     --> m-narrow-to-section
        Fixed coloring problems for numbers
        Wrote an info file to document Matlab mode 4/2/00
        Added "catch" to m-find-delimiter list (finds try) 4/5/00
        Fixed auto filling for in-line comments (fixed
          is_trailing_comment) 4/5/00
        Added m-fill-comment() 4/7/00
        Added NaN and Inf to c_number color class 5/18/00 (is_m_nan())
        Fixed is_m_nan to only allow NaN, nan, Inf, inf 6/5/00
   1.1c
        Fixed indentation after a = [ xx [xx ] etc. This involved
        repairing first_matrix. 9/15/00
        Fixed reindenting after yanking. This involved fixes to
        m-compute-indent and do-m-indent. 9/15/00
        Don't recognize end(...) to be an end statement but a call to
        an end method (get_matching_start) 9/18/00
   1.1d Matlab v6.0 support: Added "continue" to the list of
        m-keywords. 11/8/00
        Added display of function name in mode-line. 11/27/00
   1.1e Added m-transpose-lines which reindents after a transpose based
        on value of reindent-after-matlab-yank. 12/6/00
        Fixed a problem with first_continuation which got fooled when
           y = f(x) ...
        is in a string or in a comment
        Fixed a problem with comment flowing when the flowed word was
        end/else/catch etc. 12/8/00
   1.1f The problem with comment flowing was not fixed properly. Tried
        another approach. Also fixed display of function name in mode
        line when in a script - it should display nothing. 1/09/01
        Added m-left-delim which inserts a right delim and backs up
        when user inserts a left delim. Controlled by variables:
          m-auto-right-delim
          m-auto-right-delim-chars
          1/09/01
   1.1g Added m-extra-keywords to allow a list of additional keywords
        to be colored the same as matlab keywords 1/10/01
        Fixed coloring of end when used as an index 1/17/01
        Added mdoc to edoc 1/17/01
   1.1h Fixed case-sensitive handling of matlab keywords, i.e. (Else is
        a legal variable name and should not be reindented or colored)
        Added m-insert-semicolon to automatically insert a ; after an =
        sign which starts a new assignment. This can be controlled with
        the user variable m-auto-semicolon.
        Replaced leave-m-comment with a more general m-carriage-return,
        still bound to Ctrl-<enter>. This does the old leave-m-comment
        if you are in a comment, otherwise it snatches you out of the
        middle of a statement to start a new line.
        2/2/01
   1.1j Color hanging delimiters in a new color class. 3/19/01
        Color unmatched delimiters ([{ 3/21/01
        Color uncompleted strings 3/21/01
        Added Matlab's own color classes rather than use c classes.
        Introduced 3 new classes:
          m_unmatched_delim - } without a {, etc
          m_unmatched_keyword - hanging keyword: end without a for/while/if
          m_unterminated - string = 'xxx without a terminating quote
   1.1k Corrected coloring problems with distinguishing +/- as
        operators or part of a number 4/4/01
   1.1l Improved is_in_m_comment so it isn't fooled with % inside of a
        string. 4/7/01
   1.1m added m-previous-line{-back} 5/02/01
   1.1n Reworked coloring of hanging keywords and unmatched delimiters.
        For large files it didn't work to invalidate the whole buffer.
        So now we do the standard coloring and search a broadened
        region for unmatched items only. The size of this region is
        controlled by two variables.
        Minor - fixed coloring of binary operations for x'+y 6/29/01
   1.1o Removed m-fill-comment. Now simply bind Alt-q to fill-c-comment
        prior to Eps 11 and fill-comment thereafter
   1.1p Modified m-tag-case, 0-don't tag case statements, 1-tag case
        statements which appear in column 1, 2-tag all case statements
        1/31/02
   1.1q Fixed coloring of punctuation by replacing c_color_init with
        m_color_init 3/14/03
   1.1r Replaced looping through NumKeys with copy_key_table for Eps 12.
        Moved the loading of the menu from matlab-mode to
        matlab-context-menu and contrain it to only loading once. This
        was a consequence of eps12 executing a files mode when tagging.
        If tagging many functions the menu gets re-built over and over
        slowing things down tremendously.
   1.1s Eliminated all menu loading and simply set gui-menu-file =
        "tosctex.mnu"
   1.1t Added function to the list of keywords matching end in order to
        handle coloring and indenting properly - need to fix
        matlab_func_name_finder for nested functions and then it would
        be nice to add a comment after the end to indicate which
        function it matches. Added support for classdef including
        coloring and matching end for properties and methods keywords
        4/11/06
   1.1u Fixed a bug in m_compute_indent which caused flow_comment to
        add a spurious % to the start of some lines - 7/04/08
   1.1v Lugaru fixed a minor incompatibility with CUA keyboard. 02/20/12
*/


#include "eel.h"
#include "c.h"
#include "matlab.h"
#include "colcode.h"
#include "menu.h"

// function ([optional]=) (funcname)
#define MATLAB_FUNC "^function( %[?[A-Za-z0-9_, \t]+%]?[ \t]*=)?[ \t]*([A-Za-z0-9_]+)"
/* the delimiters in use */
#define LEFTD	"[({"
#define RIGHTD	"])}"
#define M_START_DELIMITERS "(switch|if|while|for|try|classdef|properties|methods)"
#define M_ENDELSE_MATCHING "(end|else|if|while|for|switch|try|function|properties|methods|classdef)"
#define M_END_MATCHING "(end|if|while|for|switch|try|function|properties|methods|classdef)"
#define M_PREVIOUS_LINE 20 // this-cmd for m-previous-line
#define M_GROUP "[ \t\f]*(if|else|elseif|while|for|switch|case|otherwise|try|catch|classdef|properties|methods)[^a-zA-Z0-9]"
#define M_CLASS_ATTRIBUTES "|SetAccess|GetAccess|Sealed|Transient|Dependent|Visible|NotifyPropertyGet|NotifyPropertySet|Abstract|External|Access|Static|"

color_class m_comment     = color_class c_comment;
color_class m_keyword     = color_class c_keyword;
color_class m_class_keyword  = color_class c_keyword;
color_class m_class_attribute  = color_class c_keyword;
color_class m_function    = color_class c_function;
color_class m_identifier  = color_class c_identifier;
color_class m_string      = color_class c_string; 
color_class m_number      = color_class c_number; 
color_class m_punctuation = color_class c_punctuation;
color_class m_unmatched_delim white on red;
color_class m_unmatched_keyword white on red;  
color_class m_unterminated white on red;

int num_prev_line;
int failed_prev_line;
char WorkString[120];

when_loading() // define m_cx_tab defaults
{
#if (EELVERSION<120)  
  int i;
#endif
  if (lookpath("tosctex.mnu"))
    strcpy(gui_menu_file,"tosctex.mnu");
  strcpy(matlab_version,MATLABVERSION); // set new version number
  // remove binding of 'f' left over from previous version
  m_tab['f'] = 0;
  // remove old names from the Name Table
  drop_name("find_next_section");
  drop_name("find_previous_section");
  drop_name("m_indent_region");
  drop_name("narrow_to_section");
  drop_name("comment_next_line");
  drop_name("is_in_comment");
  drop_name("is_in_string");
  drop_name("leave_m_comment");
  drop_name("m_fill_comments");
  drop_name("m_fill_comment");
  drop_name("is_m_start_block");
#if (EELVERSION<120)  
  for (i = 0; i < NUMKEYS; i++)
    m_cx_tab[i] = cx_tab[i];
#else
  copy_key_table(cx_tab, m_cx_tab);
#endif
}

command matlab_context_menu() on m_tab[MOUSE_RIGHT_DN], m_tab[MOUSE_DBL_RIGHT]
{
  if (lookpath("tosctex.mnu"))
  {
    win_menu_popup("_matlabpopup"); // If the file exists it was loaded
    invisible_cmd();
  }
}

// Automatic indentation for m code (Matlab)

m_indenter() // matlab indents
{
  indent_to_column(m_compute_indent());
}

command matlab_mode()
{
  fix_key_table(reg_tab, find_index("cx-tab"), m_tab, find_index("m-cx-tab"));
  mode_keys = m_tab;
  // It seems that it a better approach would be for show_matching_delimiter
  // to interogate Matchdelim. With the currrent approach you cannot turn
  // Matchdelim on or off without re-entering matlab mode.
  m_tab[']'] = Matchdelim ? (short) show_matching_delimiter : 0;
  m_tab[')'] = Matchdelim ? (short) show_matching_delimiter : 0;
  m_tab['}'] = Matchdelim ? (short) show_matching_delimiter : 0;
  indenter = m_indenter;
  auto_indent = 1;
  major_mode = _m_mode_name;
  strcpy(comment_start, "%%[ \t]*"); // check this
  strcpy(comment_pattern, "%%[ \t]*.*$");
  strcpy(comment_begin, "% ");
  strcpy(comment_end, "");
  recolor_range = color_m_range;  // set up coloring rules
  recolor_from_here = color_m_from_here;
// coloring_flags |= COLOR_INVALIDATE_FORWARD | COLOR_INVALIDATE_BACKWARD;
  if (want_code_coloring)   // maybe turn on coloring
    when_setting_want_code_coloring();
  if (m_mode_mouse_to_tag)
    mouse_goes_to_tag = 1;
#if (EELVERSION > 89)
  auto_show_matching_characters = m_auto_show_delimiters ?
                                  m_auto_show_delim_chars : "";
#endif
#if (EELVERSION<110)
  m_tab[ALT('q')] = (short) fill_c_comment;
#else
  m_tab[ALT('q')] = (short) fill_comment;
#endif
  make_mode();
}

suffix_m()
{
  matlab_mode();
}

suffix_asv()
{
  matlab_mode();
}

jmp_buf m_indent_giveup;  /* come here when we hit narrowed edge */

command do_m_indent() on m_tab['\t']
{  // This was taken from lugaru's do_c_indent()
  int orig = point,  ind = get_m_indent();
  int orig_column = current_column();
  if (maybe_indent_rigidly(0))
    return;
  to_indentation();
  if (!m_tab_always_indents && orig_column > current_column()) {
    // not in indentation
    point = orig;
    to_column(orig_column+tab_size - orig_column % tab_size);
  } else if (prev_cmd == M_INDENT) // repeated, make bigger
    to_column(orig_column + ind);
  else {
    save_var m_comment_flow = 0; // This indent should not do comment flowing
    save_var m_comment_column = 0;
    m_indenter();
    
  }
  this_cmd = M_INDENT;
}


// Augments transpose-lines() by reindenting the lines
command m_transpose_lines() on m_cx_tab[CTRL('T')]
{
  if (reindent_after_matlab_yank) {
    transpose_lines();
    save_spot point, mark;
    to_end_line();
    mark=point;
    to_begin_line();
    point--;
    to_begin_line();
    indent_region();
  }
}

/*  Look at the last two lines and return the correct indentation
    for the current line, assuming M source (or similar).
    Also reindents previous line as necessary.
    At start, we must be at the end of a line's indentation.
*/

m_compute_indent()
{
  int ind = 0;  // indentation to use
  int orig = point;
  int pstart;   // start position of prev line
  int trailing_comment;
  int keyword_indent = 0;
  int flowing_comment = 0;
  spot pcurrent = alloc_spot();
  if (setjmp(&m_indent_giveup)) // if we hit end & gave up
    return ind;                 // use best guess so far 

  // No keyword reindent when flowing in a comment
  to_begin_line();
  point--;
  keyword_indent = (!is_in_m_comment() || !m_comment_flow);
  point = orig;
  if (keyword_indent) {
    if (is_m_else(orig))
      return get_matching_start_column();
    if (is_m_case(orig))
      return get_previous_case_column();
    if (is_m_catch(orig))
      return (get_catch_column());
  }

  to_begin_line();
  mind_space();     // skip whitespace
  trailing_comment = (is_trailing_comment() &&
                      no_blank_lines(*pcurrent) &&
                      m_comment_column);

  to_m_statement_start(); // first significant character
  pstart = point;

// Do we need the first three cases now that we re-indent after the
// last character? Yes, because auto-reindent might be turned off.
  if (is_m_else(pstart)) {   // re-indent end, else, and elseif on 
    ind = get_matching_start_column(); // previous line
    to_m_statement_start();
    indent_to_column(ind);
  }
  else if (is_m_case(pstart)) { // case followed by a case or otherwise
                                // should be aligned
    ind = get_previous_case_column();
    to_m_statement_start();
    indent_to_column(ind);
  }
  else if (is_m_catch(pstart)) { // catch re-indented to previous try
    ind = get_catch_column();
    to_m_statement_start();
    indent_to_column(ind);
  }
  else if (character(pstart)=='%' && m_comment_flow &&
                        no_blank_lines(*pcurrent)) {
    if (!(indenter_action & IACT_FILL_COMMENT)
           && flow_comment(pstart,*pcurrent))
      *pcurrent = point; // adjust the return point
    ind = current_column();
    point = pstart;
  }
  else if (trailing_comment && (key != '\n')) {
    ind = get_m_comment_column();
    if (parse_string(1,".+(%%)")) {
      point = find_group(1,1);
       // double duty for flow_comment not good
      if (m_comment_flow &&
         (flowing_comment = flow_comment(point,*pcurrent))) {
        *pcurrent = point;
        ind = current_column();
        point = pstart;
      }
    }
  }
  else
    ind = current_column();

  pstart = point;

  // indentation for current line
  if (first_continuation() || first_matrix())
    ind = current_column();
  else {
    point = pstart;
    if (last_continuation())
      ind = current_column();
    else {
      point = pstart;
      if (last_matrix()) {
        if (get_assignment_start())
          ind = current_column();
      }
      else {
        if (m_statement_start()) // previous line starts a group
          if (!m_statement_ended() && !flowing_comment) // not if it is ended on same line
            ind += get_m_indent();
      }
    }
  }
    
  // reindent trailing comments on previous line
  if (m_comment_column) 
    indent_m_comment();
  point = *pcurrent;
  free_spot(pcurrent);
  return ind;
}

m_indent_alone()
{ // This assures that the indent is done and the point does not move
  save_var m_tab_always_indents = 1;
  save_spot point;
  do_m_indent();
}


indent_m_comment() // looks for next comment on line and indents it
{
  int ok_to_indent;
  if (parse_string(1,"[^%\n]+(%%)")) {
    point = find_group(1,1);
     /* this is a simple but not always correct test that the current
        % is not in a format statement
        The second condition says don't indent when multiple %'s on one line
     */
    ok_to_indent = (!is_in_m_string() && !parse_string(1,"%%.*%%"));
    if (ok_to_indent) {
      indent_to_column(get_m_comment_column());
      // Make sure there is whitespace between end of statement and %
      if (!index(" \t", character(point - 1)))
        indent_to_column(current_column()+1);
    }
  }
}

#if (EELVERSION<120)  
m_carriage_return() on m_tab[CTRL('J')]
#else
m_carriage_return() on m_tab[CTRL(GREYENTER)]
#endif
{
  int ind;
  if (is_in_m_comment()) {
    save_var m_comment_delete_empty = 0;
    ind = current_column();
    enter_key();
    indent_to_column(ind);
  } else {
    to_end_line();
    enter_key();
  }
}

// assume first line of multi-line statement has an equal sign
int first_continuation()
{
  int psave = point;
  if (parse_string(1,".*=[ \t]*([^ ].*%.%.%.)")) {
    point = find_group(1,1);
    if (!is_in_m_code()) {
      point = psave;
      return 0;
    } else
      point += parse_string(1,"([a-zA-Z0-9_]*[ ]*%(|%[)");
    return 1;
  }
  return 0;
}

// Opening line of a matrix assignment, which is not completed and
// not continued with ..., i.e. A = [1 2 3
// We assume here that we are NOT on the first line of a continuation
// Does not deal with embedded
int first_matrix()
{
  if (parse_string(1,".*=[ \t]*(%[)[^][\n]*(%[[^][\n]*%])*[^]\n]*\n")) {
    point = find_group(1,1)+1;
    if (parse_string(1,"[ \t]+(.)")) // move forward past any spaces
      point = find_group(1,1);
		return 1;
	}
  return 0;
}


int last_matrix()
{
  int num_left=0, num_right=0, psave=point;
  // count active [
  while (parse_string(1,".*%[")) {
    point = matchend;
    if ( is_in_m_code() )
      num_left++;
  }
  point = psave;
  // count active ]
  while (parse_string(1,".*%]")) {
    point = matchend;
    if ( is_in_m_code() )
      num_right++;
  }
  point = psave;
  return num_right>num_left;
}

// previous line finished a multi-line set
int last_continuation()
{
  int doindent = 0;
  if (!parse_string(1,".*%.%.%.")) { // this line is not continued
    while (m_is_contin()) {
      point--;
      to_begin_line();
      doindent = 1;
    }
    if (doindent) 
      to_indentation();
  }
  return doindent;
}
     
mind_space()    /* move back over any whitespace or comments */
{     /* give up if we hit start */
  if (mind_try_space())
    return;  // haven't really correctly implemented this longjmp yet
//    longjmp(&m_indent_giveup, 1);
}

mind_try_space()  // move back over whitespace, comments
{                 // lines, and return 0, or 1 if we hit start
  for (;;) {
    if (index(" \t\n\f", character(point - 1)))
      point--;
    else
      return point == narrow_start;
  }
}

int is_m_else(p) // is there an else, elseif, or end keyword at point
int p;
{
  save_var point=p, case_fold = 0;
  return parse_string(1, "(else|elseif|end)([^a-zA-Z0-9_%(]|$)");
}

int is_m_case(p) // assume at start of possible case or otherwise
int p;
{
  save_var point = p, case_fold = 0;
  return parse_string(1, "(case|otherwise)([^a-zA-Z0-9_]|$)");
}

int is_m_catch(p) // assume at start of possible case or otherwise
int p;
{
  save_var point = p, case_fold = 0;
  return parse_string(1, "catch([^a-zA-Z0-9_]|$)");
}

int get_previous_case_column()
{
  int col, col0 = current_column();
  save_var point, case_fold = 0;
  while (re_search(-1,"(end|case|switch)[^a-zA-Z0-9_]")) {
    if (!parse_string(-1,"[a-zA-Z_]"))
      if (is_in_m_code() && !is_end_index()) { // it is a statement
        if (character(point) == 'e') {// found an end move past its start
          if (!get_matching_start(0))
            break;
        } else {
          col = current_column();
          if (character(point) == 's') // found switch so indent
            col += get_m_case_indent();
          return col;
        }
    }
  }
  return col0;
}

int get_catch_column()
{
  int col0 = current_column();
  save_var point, case_fold = 0;
  while (re_search(-1,"(end|try)[^a-zA-Z0-9_]")) {
    if (!parse_string(-1,"[a-zA-Z_]"))
      if (is_in_m_code() && !is_end_index()) { // it is a statement
        if (character(point) == 'e') {// found an end move past its start
          if (!get_matching_start(0))
            break;
        } else
            return current_column();
    }
  }
  return col0;
}

/* Checks if the previous statement is an end */
int m_statement_ended()
{
  int psave = point;
  save_var point, case_fold = 0;
  if (parse_string(1,"(.+)%%")) // go to beginning of trailing comment
    point = find_group(1,0);
  else {
    point = psave;
    nl_forward();
    point--;
  }
  return parse_string(-1,"[^a-zA-Z0-9_]end[ \t\f,;]*");
}
   
to_m_statement_start()
{
  to_begin_line();
  to_indentation();
}
/*
Tell if the last line begins a statement (rather than a function
declaration or the continuation of a previous statement).
Assumes point is at the start of the last line.
*/
m_statement_start()
{
  save_var case_fold = 0;
  return parse_string(1, M_GROUP);
}

int get_assignment_start()
{
  return re_search(-1,"[a-zA-Z][a-zA-Z0-9_]*[ \t]*=");
}

int get_matching_start_column()
{
  save_var point;
  get_matching_start(character(point+1) == 'l'); // match else or end
  return current_column(); // if not found it will return current column
}

/* match else, end, or catch to their if,while,for,try or switch
   assumes point is just before end, else or catch */
int get_matching_start(iselse)
// if a 1 is returned it matches a starting keyword (for, if, while,...
// if a 2 is returned it matches a starting class keyword (function,
//     classdef, properties, methods
// Be careful in the future because we only check single characters in
// the match
int iselse; // ==1 matching else, ==2 matching catch
{
  int extra_ends = 0;
  int ret;
  if (iselse<0)
    return 1;
  // Also must check for embedded ends
	while (do_searching(REGEX|WORD|REVERSE,M_ENDELSE_MATCHING)) {
		if ((is_in_m_code() && !is_end_index()) &&
             !is_structure_field()) {
			if (character(point) == 'e') { // end or else but not end(..
				if ( (character(point+1) == 'n') && (character(point+3) != '(') ) 
					extra_ends++;
				else // if an else was entered and the first find was an else there is a error
					if ((iselse==1) && !extra_ends) 
						return 0;
			}
            else if (--extra_ends<0) { // match found: is it correct
              switch (iselse) {
                case 0: ret = keyword_type(point); break;
                case 1: ret = (character(point)=='i'); break;  // an else must be matched with an if
                case 2: ret = (character(point)=='t'); break; // a catch must be matched with a try
              }
              return ret;
            }
		}
  }
  return 0;
}

int keyword_type(position) // this assumes that a keyword is at postion
{ // a 1 is returned for standard keyword, a 2 for class keywork including function
  return ((character(position)=='c') || (character(position)=='p') ||
          (character(position)=='m') ) + 1;
//  (character(position)=='m') || (character(position+1)=='u')) + 1;
}

// point is assumed to be at start of keyword so we initialize extra_ends to -1
int get_matching_finish()
{
  int extra_ends = -1;
  while (do_searching(REGEX|WORD,M_END_MATCHING)) {
    if (is_in_m_code() && !is_structure_field()) {
      if (character(point-1) == 'd') { // end but not end(.. -- overload of end
        if ( !parse_string(1,"[ \t]*%(") && !is_end_index_at(point-3) )
          if (!extra_ends)
            return 1;
          else
            extra_ends--;
      }
        else // we found another block start
          extra_ends++;
    }
  }
  return 0;
}

/* is end used as an index rather than a group end, could be an array index
   or a cell array index */
int is_end_index()
{
  int isndx = 0, psave = point, start;
  save_var point, case_fold = 0;
  if (!parse_string(1,"end"))
    return 0;
  to_m_statement_start();
  start = point;
  point = psave;
  while (re_search(-1,"(%(|%))")) { // ordinary array
    if (point < start) // no need to search beyond start of statement
      break;
    if (character(point) == ')')
      isndx++;
    else if (--isndx<0)
      return 1;
  }
  point = psave;
  while (re_search(-1,"({|})")) { // cell array
    if (point < start)
      break;
    if (character(point) == '}')
      isndx++;
    else if (--isndx<0)
      return 1;
  }
  return 0;
}

// Attempts to match end statement
// Removed binding to 'f', This worked but caused elseif to find the if twice
// once for the 'e' and once for the 'f'
// renamed from m_show_matching_delimiter
command m_reindent_keyword() on m_tab['d'], m_tab['e'], m_tab['h'] //, m_tab['f']
{
  save_var case_fold = 0;
  normal_character();
  if (is_in_m_code() && !is_end_index_at(point-3)) {
                                     // don't match x.end etc
    if (m_auto_reindent && parse_string(-1,"[^%.](end|else|case|otherwise|catch)")
                        && (parse_string(1,"<wspace>") || point==size())) {
      m_indent_alone();
      if (m_matchdelim)
        m_find_delimiter();
		}
  }
}

command m_find_delimiter() on m_tab[ALT(']')]
// Alt-] will match elseif but we have unbound it from m-show-matching-delimiter
{
  char buf[7];
  int iselse;
  save_var point, case_fold = 0;
  
  if (parse_string(-1,"[^a-zA-Z0-9_%.]+(end|else|elseif|catch)[^a-zA-Z0-9_]*")) {
#if (EELVERSION > 99) 
    grab(point=find_group(1,1),find_group(1,0),buf);
#else
    grab(point=MIN(find_group(1,0),find_group(1,1)),
         MAX(find_group(1,0),find_group(1,1)),buf);
#endif
    if (is_end_index())
      return;
    else
      switch (character(point+1)) {  // identify special cases
        case 'l': iselse = 1; break; // else and elseif
        case 'a': iselse = 2; break; // catch
        default:  iselse = 0; break; // end
      }
    say("");
    iter=0;
    if (in_macro())
      return;
    if (get_matching_start(iselse))
      show_line();
    else {
      maybe_ding(m_want_bell);
      say("Unmatched %s statement",buf);
    }
  }
}

m_spacebar() on m_tab[' ']
{
  if (is_in_m_comment())
    m_comment_handling();
  else
    m_insert_semicolon();
}

// insert a ';' following a '= '
m_insert_semicolon()
{
  if (m_auto_semicolon && parse_string(-1,"[^~=><]=") &&
      ((point == size()-narrow_end) || parse_string(1,"[ \t]*\n")))
  { // check if this is a function line or a for loop, then no ;
    save_var narrow_start = give_begin_line();
    if (!parse_string(-1,"function .+") &&
        !parse_string(-1,"for +[a-zA-Z0-9_]* +=")) {
      insert(';');
      point--;
    }
  }
  normal_character();
}  

command m_previous_line_back() on m_tab[CTRL('n')]
{
  if (prev_cmd == M_PREVIOUS_LINE) {
    has_arg = 1;
    iter = -1;
    m_previous_line();
  } else {
    maybe_ding(m_want_bell);
    say("m-previous-line-back must follow m-previous-line or back");
  }
}

command m_previous_line() on m_tab[CTRL('p')]
{
  if (!has_arg)
    iter = 1;
  if (prev_cmd==M_PREVIOUS_LINE) {
    num_prev_line += iter;
    if (!failed_prev_line) {
      point++;
      kill_region();
    }
  } else
    num_prev_line = iter;
  if (num_prev_line<0)
    num_prev_line = 0;
  if (failed_prev_line = !stuff_previous_line(num_prev_line)) {
    if (num_prev_line>0)
      num_prev_line--;
  }
  say("num = %d",num_prev_line);
  this_cmd = M_PREVIOUS_LINE;
  iter = 0;
}

int stuff_previous_line(int nline)
{
  int save=point, start;
  char line[200];
  if (nline<1) {
    maybe_ding(m_want_bell);
    say("No more lines");
    return 0;
  }
  to_begin_line();
  while (nline-- && point>0) {
    point--;
    to_begin_line();
    if (character(point)=='\n') // empty line
      nline++;
  }
  if (nline>0) {
    point = save;
    maybe_ding(m_want_bell);
    say("No more lines");
    return 0;
  } else {
    start = point;
    to_end_line();
    grab(start,point,line);
    point = save;
    to_begin_line();
    mark = point;
    bprintf("%s\n",line);
    point--;
  }
  return 1;
}  

m_comment_handling()
{
  int broke = 0;
  spot old;
  if (is_trailing_comment() && !m_inline_ident) {
    normal_character();
    return;
  }
  if (m_auto_fill) {
    old = alloc_spot();
    if (m_comment_column && is_trailing_comment()) {
      if (character(point-1)=='%') {
          to_begin_line();
          indent_m_comment();
          point = *old;
      } else
        broke = m_maybe_break_line();
    }
    else if (is_in_m_comment())
      broke = m_maybe_break_line();
    free_spot(old);
    if (broke && ((character(point-1)==' ') || (character(point-1)=='\t')) )
      return;
  }
  normal_character();
}

// Is point inside of trailing comment
// trailing means some text between ^ and %
is_trailing_comment()
{
  int beg, psave=point;
  save_var point;
  to_begin_line();
  beg = point; point = psave;
  search(-1,"%");
  while (point>beg) {
    if (is_in_m_code())
      return parse_string(-1,"^[ \t\f]*[a-zA-Z0-9]+.+");
    else
      search(-1,"%");
  }
  return 0;
}
    
  
// this was swiped from format.e and modified.
m_maybe_break_line()
{
  int h, found = 1;
  spot old;

  if (m_auto_fill && is_in_m_comment()) {
    h = current_column() + 1;
    if (h == margin_right) {
        break_line_here();
        return 1;
    } else if (h > margin_right) {
      old = alloc_spot();
      found = pick_break(margin_right);
      if (found) {
        break_line_here();
        key = ' ';
        point = *old;
      } else
        break_line_here();
      free_spot(old);
      return 1;
    }
    return 0;
  }
  return 0;
}

no_blank_lines(there) /* checks for blank lines between point and there */
int there;
{
  save_var point;
  if (re_search(1,"\n[ \t]*\n"))
    return point > there;
  return 1;
}
  
int flow_comment(here,there)
int here, there;
{
  int col0, col;
  point = here;
  col0 = current_column();
  if (parse_string(1,"%%[ \t]*")) // single % followed by whitespace
    point = find_group(0,0);
  // end of line => empty comment, delete it
  if ((character(point) == '\n') && m_comment_delete_empty) { 
    delete(find_group(0,1),point);
    return 0;
  } else {
    col = current_column();
    point = there;
    indent_to_column(col0);
    insert('%');
    indent_to_column(col);
    return 1;
  }
}
  
// This doesn't deal with format statements
// it does now
int is_in_m_comment()
{
  int psave = point;
  save_var point;
  to_begin_line();
  while ( (re_search(1,"%%")) && (point<=psave) )
    if (!is_in_m_string())
      return 1;
  return 0;
}

int is_in_m_string()
{
  int psave = point;
  save_var point;
  to_begin_line();
  while ( (re_search(1,"\'")) && (point<=psave) )
    if (definitely_not_transpose(point-1)) {
      find_right_quote();
      if (point>psave)
        return 1;
    }
  return 0;
}

is_in_m_code()
{
  return (!is_in_m_comment() && !is_in_m_string());
}

int is_structure_field()
{
  int test = parse_string(-1,"[a-zA-Z_][0-9]*%.");
  return(test);
}
   
get_m_indent() // indent this many spaces for each level
{              // if m_indent is 0, use tab_size instead
  return (m_indent > 0) ? m_indent : tab_size;
}

get_m_case_indent() // indent level for case. If < takes m_indent
{
  return (m_case_indent < 0) ? get_m_indent() : m_case_indent;
}

get_m_comment_column()
{
  return (m_comment_column > 0) ? m_comment_column : comment_column;
}

//
// Matlab Coloring we make use of many c coloring functions
//

color_m_from_here(safe) // Move backward to the nearest line guaranteed
{     // to start outside any colored region, return pos.
  to_begin_line(); // begin line should be safe for m
  while (m_is_contin()) {
    point--;
    to_begin_line();
  }
  if (point<safe)
    point = safe;
  return point;
}

/* is the current line a continuation of the previous
   it is assumed that comments can be at the end of continuation
   lines but not on a line alone although they can in certain
   cases */
m_is_contin()
{
  save_var point;
  if (nl_reverse()) {
    point -= parse_string(-1,"%%.*"); // move to start of possible comment
    return parse_string(-1,"%.%.%.[ \t]*");
  }
  return 0;
}

m_init_color(from, to)
{
	if (from < to)
		set_character_color(from, to, minimal_coloring
				    ? color_class m_identifier
				    : color_class m_punctuation);
}

color_m_range(from, to) // recolor just this section
{     // last colored region may go past to
  int t = -1, talk;
  char pat[200];

  if (from >= to)
    return to;
//  say("from: %d to: %d",from,to);
  m_init_color(from, to);
  point = from;
  talk = (to - from > 2000);  // show status during long delays
  save_var case_fold = 0;
  strcpy(pat, "%%|\'");   // start of comment or quote
  if (!minimal_coloring)
    strcat(pat, "|[A-Za-z][A-Za-z0-9_]*" // identifier
           "|[-+]" // operator or sign
           "|(%.?[0-9])[0-9]*(%.)?[0-9]*([DEde][0-9-+])?");// numbers
  while (point < to) {
    if (!re_search(1, pat)) {
      t = size();
      break;
    }
    t = matchstart;
    switch (character(point - 1)) {   // check last char
    case '%':   // found comment
      while (nl_forward() && character(point) == '%')
        ;   // Continued // comment.
      set_character_color(t, point, color_class m_comment);
      break;
    case '+':  case '-':
      if (!is_binary_op(t))
        set_character_color(t,point,color_class m_number);
      break;
    case '\'':  // found a string literal or a transpose
      if ( definitely_not_transpose(t) ) { // we are at a left quote
        if (!m_color_unterminated_strings || find_right_quote())
          set_character_color(t, point, color_class m_string);
        else
          set_character_color(t, point, color_class m_unterminated);
      } else    // we are at a transpose operator
        set_character_color(t, point, color_class m_punctuation);
      break;
    default:    // found identifier, kywd, or number
//      say("from: %d to: %d",t,point);
      set_character_color(t, point, m_keyword_color(t));
      break;
    }
    if (talk)
      note("Coloring M program: %d%% complete...",
           (point - from) * 100 / (to - from));
  }
  // pick up coloring of unmatched keywords
  if (m_color_hanging_end_delimiters)
    m_hanging_end_color(from - m_color_hanging_key_extent,
                        to + m_color_hanging_key_extent);
  if (m_color_hanging_start_delimiters)
    m_hanging_start_color(from - m_color_hanging_key_extent,
                        to + m_color_hanging_key_extent);
  if (m_color_unmatched_delimiters)
    m_hanging_delim_color(from - m_color_hanging_del_extent,
                          to + m_color_hanging_del_extent);
//  c_init_color(to, t);
  if (talk)
    note("");
  return point;
}

int is_binary_op(t) // character at t is + or -
int t;
{
  save_var point = t;
  if (!parse_string(1,"[-+]"))
    return 0;
  if (is_in_bracket())
    return ((character(t+1)==' ') || (character(t+1)=='\t')) 
        || ((character(t-1)!=' ') && (character(t-1)!='\t')
        &&  (character(t-1)!='['));
  else
    return (parse_string(-1,"[]a-zA-Z_0-9%)'][ \t]*"));
}

int is_in_bracket()
{
  return parse_string(-1,"%[[^]]*");
}
  
int definitely_not_transpose(t)
int t;
{
  return index(" -%+=,:;%%&%^%<>%|%[%({\n\t/\\", character(t-1)) ? 1 : 0;
}

int find_right_quote() // This assumes you are in a string
{
  save_var narrow_end = size() - give_end_line();
  if (re_search(1,"[^\'\n]*\'"))
    return skip_quotes() ? 1 : find_right_quote();
  return 0;
}

int skip_quotes()
{
  int i = 1;
  while(character(point)=='\'') {
    point++;
    i++;
  }
  return i % 2;
}

m_hanging_end_color(from, to)
{
  int matchndx, start, save;
  from = (from>0) ? from : 0;
  to   = (to>size()) ? size() : to;
  save_var case_fold = 0;
  save_var point = to;
  while (do_searching(WORD|REVERSE,"end") && (point>=from)) {
    start = matchstart;
    if (is_in_m_code() && !is_end_index() &&
        (character(point-1) != '.') && !paren_follows_at(point+3)) {         
      save = point;
      switch (get_matching_start(0)) {
        case 0: set_character_color(save, start, color_class m_unmatched_keyword);
                break;
        case 1: set_character_color(save, start, color_class m_keyword);
                break;
        case 2: set_character_color(save, start, color_class m_class_keyword);
      }
    }
  }
  point = to;
  while (do_searching(REVERSE|REGEX|WORD,"(else|catch)") && (point>=from)) {
    start = matchstart;
    if (is_in_m_code() && (character(point-1) != '.')) {
      matchndx = (character(point+1) == 'l') ? 1 : 2;
      save = point;
      if (!get_matching_start(matchndx))
        set_character_color(save, start, color_class m_unmatched_keyword);
      else
        set_character_color(save, start, color_class m_keyword);
      point = save;
    }
  }
}

m_hanging_start_color(from, to)
{
  int start, save;
  char buf[32];
  from = (from>0) ? from : 0;
  to   = (to>size()) ? size() : to;
  save_var case_fold = 0;
  save_var point = from;
  while (do_searching(REGEX|WORD,M_START_DELIMITERS) && (point<=to)) {
    start = matchstart;
    if (is_in_m_code() && (character(start-1) != '.')) {
      save  = point;
      point = start;
      if (!get_matching_finish())
        set_character_color(save, start, color_class m_unmatched_keyword);
      else {
        buf[0] = '|';
        grab(start, save, buf+1);
        strcpy(buf + save - start + 1, "|");
        if (is_m_class_keyword(buf))
          set_character_color(save, start, color_class m_class_keyword);
        else
          set_character_color(save, start, color_class m_keyword);
      }
      point = save;
    }
  }
}

m_hanging_delim_color(from, to)
{
  int save;
  from = (from>0) ? from : 0;
  to   = (to>size()) ? size() : to;
  save_var point = from;
  while (re_search(1,"[[{(]") && (point<=to)) {
    if (is_in_m_code()) {
      save  = point;
      point--;
      if (!move_level(1, LEFTD, RIGHTD, 0, 0))
        set_character_color(save-1, save, color_class m_unmatched_delim);
      else
        m_init_color(save-1, save);
      point = save;
    }
  }
  point = to;
  while (re_search(-1,"[]})]") && (point>=from)) {
    if (is_in_m_code()) {
      save  = point;
      point++;
      if (!move_level(-1, RIGHTD, LEFTD, 0, 0))
        set_character_color(save, save+1, color_class m_unmatched_delim);
      else
        m_init_color(save, save+1);
      point = save;
    }
  }
}

m_keyword_color(from) // return color for "identifier" from here to point
{     // (something with alpha or digits)
  char buf[500];
  int save;
//  char buf[200];

  save_var point;
  if (point - from > sizeof(buf) - 10)
    point = from + sizeof(buf) - 10;
  buf[0] = '|';   // get identifier, between | chars
  grab(from, point, buf + 1);
  case_fold = 1;
  if (index("0123456789-.", buf[1]))
    return color_class m_number;
  strcpy(buf + point - from + 1, "|");
  if (is_m_nan(buf) || is_end_index_at(from))
    return color_class m_number;
  if (is_m_keyword(buf) && character(from-1) != '.') {
    if (strstr("|end|",buf)) {
      save = point;
      point = from;
      if (get_matching_start(0)==2)
        return color_class m_class_keyword;
      else
        return color_class m_keyword;
      point = save;
    } else
      return color_class m_keyword;
  }
  if (is_m_class_keyword(buf) && character(from-1) != '.')
    return color_class m_class_keyword;
  if (color_class m_function != color_class m_identifier
      && paren_follows())
    return color_class m_function;
  if (is_m_class_attribute(buf) && character(from-1) != '.')
    return color_class m_class_attribute;
  return color_class m_identifier;
}

paren_follows_at(p)
{
  save_var point = p;
  return paren_follows();
}

is_end_index_at(p)
{
  save_var point = p;
  return is_end_index();
}


is_m_nan(char *p)
{
  save_var case_fold = 0;
  return (strstr("|inf|Inf|nan|NaN|",p)) ? 1 : 0;
}

is_m_keyword(p)   // is text in p (must be surrounded by |'s) a keyword?
char *p;
{
  save_var case_fold = 0;
  if (strstr("|if|else|elseif|end|while|for|return|break|switch|"
       "case|otherwise|global|function|try|catch|"
       "persistent|continue|",p) ||
      strstr(m_extra_keywords,p))
    return 1;
  return 0;
}

is_m_class_keyword(p)   // is text in p (must be surrounded by |'s) a keyword?
char *p;
{
  save_var case_fold = 0;
  if (strstr("|classdef|properties|methods|",p))
    return 1;
  return 0;
}


is_m_class_attribute(p)   // is text in p (must be surrounded by |'s) a keyword?
char *p;
{
  save_var case_fold = 0;
  if (strstr(M_CLASS_ATTRIBUTES,p))
    return 1;
  return 0;
}

//
// Matlab Code for tags and jumping
// 

// matlab mode uses the default pattern for tagging
// tag-pattern-default
tag_suffix_m()
{
  char func[TAGLEN];
  char fname[FNAMELEN], *p;
  int start, next;
  save_var case_fold = 0;
  if (m_add_objdirs_to_tag && m_tag_function) {
    strcpy(fname,filename);
    p = get_tail(fname,0);
    p--;
    *p = '\0';
    p = get_tail(fname,0);
    if (p[0]=='@')
      p++;
    else
      *p = '\0';
  }
  while (find_elseif_section(1)) {
    grab(start = find_group(1, 1), find_group(1, 0), func);
    add_tag(func, start);
  }
  point = 0;
  while (find_case_section(1)) {
    grab(start = find_group(1, 1), next = find_group(1, 0), func);
    add_tag(func, start);
		point = next+1;
		while (parse_string(1,"[ \t,]+'([^'\n]+)'")) {
			grab(start = find_group(1, 1), next = find_group(1, 0), func);
			add_tag(func, start);
			point = next+1;
		}
  }
  point = 0;
  while (find_function_section(1)) {
    grab(start = find_group(2, 1), find_group(2, 0), func);
    if (m_add_objdirs_to_tag && strlen(p)>0)
      sprintf(func,"%s: %s",func,p);
    add_tag(func, start);
  }
}

command m_find_next_section() on m_tab[ALT('.')]
{
  int temp = point;
  nl_forward();
  if (!find_section(1)) {
    point = temp;
    say("No more sections.");
  } else
    say("");
}

command m_find_previous_section() on m_tab[ALT(',')]
{
  int temp = point;
  nl_reverse();
  if (!find_section(-1)) {
    point = temp;
    say("No more sections.");
  } else
    say("");
}

int find_section(dir)
int dir;
{
  int pe, pstart=point;
  char rtn=0;
  save_var case_fold = 0;
  if (dir>0)
    pe = size();
  else
    pe = 0;
  if (find_elseif_section(dir)) {
    rtn = 1;
    if (dir>0)
      pe = matchstart;
    else
      pe = point;
  } else
  point = pstart;
  if (find_case_section(dir)) {
    rtn = 1;
    if (dir>0)
      pe = MIN(matchstart,pe);
    else
      pe = MAX(point,pe);
  }
  point = pstart;
  if (find_function_section(dir)) {
    rtn = 1;
    if (dir>0)
      pe = MIN(matchstart,pe);
    else
      pe = MAX(point,pe);
  }
  if (rtn)
    point = pe;
  else
    point = pstart;
  return rtn;
}

int find_elseif_section(dir)
int dir;
{
  save_var case_fold = 0;
  if (!m_tag_elseif)
    return 0;
  return re_search(dir,"^elseif.+'([^'\n]+)'");
}

int find_case_section(dir)
int dir;
{
  save_var case_fold = 0;
  switch (m_tag_case) {
    case 1:  return re_search(dir,"^case[ ]+{*[ \t]*'([^'\n]+)'");
    case 2:  return re_search(dir,"^[ ]*case[ \t]+{*[ \t]*'([^'\n]+)'");
    default: return 0;
  }
}

int find_function_section(dir)
int dir;
{
  if (!m_tag_function)
    return 0;
  return re_search(dir,MATLAB_FUNC);
}

matlab_func_name_finder() // must be updated for nested functions - tjb
{
  save_var point;
  to_end_line();
  if (re_search(-1,MATLAB_FUNC)) 
    grab(find_group(2,1),find_group(2,0),display_func_name);
  else
    *display_func_name = 0;
  return 1;
}
    
command m_narrow_to_section() on m_tab[ALT('n')]
{
  if ( (narrow_start>0) || (narrow_end>0) ) 
    widen_buffer();
  else {
    save_var point;
    point++;
    if (!find_section(1))
      point = size();
    set_mark();
    if (find_section(-1)) {
      to_begin_line();
      narrow_to_region();
    }
  }
}

command matlab_show_version()
{
  say("Matlab Mode %s", matlab_version);
}

// ------------ commenting lines --------------------------

int m_comment_line(int onoff)
{ // puts a comment on current line (or takes it off)
  save_spot point;
  to_begin_line();
  if (onoff>0) {
    insert('%');
    return 1;
  } else if (parse_string(1,"[ \t]*%%")) {
    delete(matchend-1,matchend);
    return 1;
  }
  return 0;
}

m_line_comment(int onoff)
{
  int save = point;
  to_end_line();
  if (point<size()) {
    point++;
    if (!m_comment_line(onoff))
      point = save;
  } else
    point = save;
}


command m_comment_next_line() on m_cx_tab['%']
{
  if (!has_arg && is_highlight_on())
    m_region_comment(1);
  else
    m_line_comment(1);
}

command m_uncomment_next_line() on m_cx_tab['$']
{
  if (!has_arg && is_highlight_on())
    m_region_comment(-1);
  else
    m_line_comment(-1);
}

m_region_comment(int onoff)
{
  int save = mark;
  save_spot point, mark;
  if (point>mark) {
    mark = point;
    point = save;
  }
  while (point<mark) {
    m_comment_line(onoff);
    to_end_line();
    if (point<size())
      point++;
    else
      break;
  }
} 

/* Might want to implement this in the future
matlab_find_linked_file(char *file)
{
  say("Matlab find linked file");
}*/

command m_comment_region() on m_cx_tab[FKEY(4)]
{
  m_region_comment(1);
}

command m_uncomment_region() on m_cx_tab[FKEY(3)]
{
  m_region_comment(-1);
}

// insert right delim following a left
command m_left_delim() on m_tab['('], m_tab['['], m_tab['{'], m_tab['\'']
{
  if (!m_auto_right_delim) {
    normal_character();
    return;
  }
  iter = 0;
  if (point < size() - narrow_end && !strchr(" ;)}]\t\n}", curchar()))
    normal_character();
  else {
    if (strchr(m_auto_right_delim_chars,key)) {
      switch (key) {
        case '(': stuff("()"); break;
        case '[': stuff("[]"); break;
        case '{': stuff("{}"); break;
        case '\'':
          if (strchr(" =,;[({\n\t", character(point-1))) {
            stuff("''");
            break;
          } else {
            normal_character();
            return;
          }
      }
      point--;
    } else
      normal_character();
  }
}
